@@ -46,6 +46,7 @@
 #endif /* USE_HAL */
 
 #ifdef __linux__
+#include <linux/cdrom.h>
 #include <scsi/scsi.h>
 #include <scsi/sg.h>
 #endif /* __linux__ */
@@ -58,6 +59,8 @@
 
 #include "cd-drive.h"
 
+#define CD_ROM_SPEED 176
+
 static struct {
 	const char *name;
 	gboolean can_write_cdr;
@@ -69,8 +72,21 @@ static struct {
 	{ "SONY - CD-R   CDU948S",  TRUE, FALSE, FALSE, FALSE },
 };
 
+typedef enum {
+	CD_PROTOCOL_IDE,
+	CD_PROTOCOL_SCSI,
+} CDProtocolType;
+
+struct CDDrivePriv {
+	CDProtocolType protocol;
+	char *udi;
+};
+
+static CDDrive *cd_drive_new (void);
+
 /* Utility functions, be careful to have a match with what's use in the
  * different bits of code */
+#if !defined (__linux__)
 static char *
 cdrecord_get_stdout_for_id (char *id)
 {
@@ -104,9 +120,8 @@ cdrecord_get_stdout_for_id (char *id)
 	return NULL;
 }
 
-#if !defined (__linux__)
 static void
-get_cd_properties (char *id, int *max_rd_speed, int *max_wr_speed,
+get_cd_properties (char *device, char *id, int *max_rd_speed, int *max_wr_speed,
 	CDDriveType *type)
 {
 	char *stdout_data, *rd_speed, *wr_speed, *drive_cap;
@@ -115,54 +130,13 @@ get_cd_properties (char *id, int *max_rd_speed, int *max_wr_speed,
 	*max_wr_speed = -1;
 	*type = 0;
 
+	*max_rd_speed = get_device_max_read_speed (device);
+	*max_wr_speed = get_device_max_write_speed (device);
+
 	stdout_data = cdrecord_get_stdout_for_id (id);
 	if (stdout_data == NULL) {
 		return;
 	}
-
-	wr_speed = strstr (stdout_data, "Maximum write speed:");
-	if (wr_speed != NULL) {
-		char *tok;
-		wr_speed += strlen ("Maximum write speed:");
-		for (tok = strtok (wr_speed, " (),\t\n");
-				tok && strcmp (tok, "CD");
-				tok = strtok (NULL, " (),\t\n")) {}
-		tok = strtok (NULL, " (),\t\n"); /* Get the CD speed. */
-		*max_wr_speed = atol (tok);
-	} else {
-		wr_speed = strstr (stdout_data, "Maximum write speed in kB/s:");
-		if (wr_speed != NULL) {
-			char *tok;
-			wr_speed += strlen ("Maximum write speed in kB/s:");
-			for (tok = strtok (wr_speed, " (),\t\n");
-					tok && strcmp (tok, "CD");
-					tok = strtok (NULL, " (),\t\n")) {}
-			tok = strtok (NULL, " (),\t\n"); /* Get the CD speed. */
-			*max_wr_speed = atol (tok);
-		}
-	}
-
-	rd_speed = strstr (stdout_data, "Maximum read  speed:");
-	if (rd_speed != NULL) {
-		char *tok;
-		rd_speed += strlen ("Maximum read  speed:");
-		for (tok = strtok (rd_speed, " (),\t\n"); 
-				tok && strcmp (tok, "CD"); 
-				tok = strtok (NULL, " (),\t\n")) {}
-		tok = strtok (NULL, " (),\t\n"); /* Get the CD speed. */
-		*max_rd_speed = atol (tok);
-	} else {
-		rd_speed = strstr (stdout_data, "Maximum read  speed in kB/s:");
-		if (rd_speed != NULL) {
-			char *tok;
-			rd_speed += strlen ("Maximum read  speed in kB/s:");
-			for (tok = strtok (rd_speed, " (),\t\n");
-					tok && strcmp (tok, "CD");
-					tok = strtok (NULL, " (),\t\n")) {}
-			tok = strtok (NULL, " (),\t\n"); /* Get the CD speed. */
-			*max_rd_speed = atol (tok);
-		}
-	}
 	drive_cap = strstr (stdout_data, "Does write DVD-RAM media");
 	if (drive_cap != NULL) {
 		*type |= CDDRIVE_TYPE_DVD_RAM_RECORDER;
@@ -219,6 +193,10 @@ add_whitelist (CDDrive *cdrom)
 /* For dvd_plus_rw_utils.cpp */
 int get_dvd_r_rw_profile (const char *name);
 int get_mmc_profile (int fd);
+int get_disc_size_cd (int fd);
+gint64 get_disc_size_dvd (int fd, int mmc_profile);
+int get_read_write_speed (int fd, int *read_speed, int *write_speed);
+
 
 static void
 add_dvd_plus (CDDrive *cdrom)
@@ -241,6 +219,30 @@ add_dvd_plus (CDDrive *cdrom)
 	}
 }
 
+static gboolean
+cd_drive_door_open (gboolean mmc_profile, int fd)
+{
+#ifdef __linux__
+	{
+		int status;
+
+		status = ioctl (fd, CDROM_DRIVE_STATUS, CDSL_CURRENT);
+		close (fd);
+		if (status < 0) {
+			return FALSE;
+		}
+
+		return status == CDS_TRAY_OPEN;
+	}
+#else
+	if (mmc_profile == 0) {
+		return TRUE;
+	}
+
+	return FALSE;
+#endif
+}
+
 CDMediaType
 cd_drive_get_media_type_from_path (const char *device)
 {
@@ -249,6 +251,8 @@ cd_drive_get_media_type_from_path (const char *device)
 
 	g_return_val_if_fail (device != NULL, CD_MEDIA_TYPE_ERROR);
 
+	/* FIXME get media type from HAL if it's there BZ 145196 */
+
 	fd = open (device, O_RDONLY|O_EXCL|O_NONBLOCK);
 	if (fd < 0) {
 		if (errno == EBUSY) {
@@ -258,11 +262,26 @@ cd_drive_get_media_type_from_path (const char *device)
 	}
 
 	mmc_profile = get_mmc_profile (fd);
+
+	/* Couldn't get the data about the media */
+	if (mmc_profile <= 0) {
+		gboolean opened;
+
+		opened = cd_drive_door_open (mmc_profile, fd);
+		close (fd);
+
+		if (opened != FALSE) {
+			return CD_MEDIA_TYPE_ERROR;
+		} else {
+			return CD_MEDIA_TYPE_UNKNOWN;
+		}
+	}
+
 	close (fd);
 
 	switch (mmc_profile) {
-	case -1:        /* Couldn't get the data about the media */
-		return CD_MEDIA_TYPE_UNKNOWN;
+	case -1:
+		g_assert_not_reached ();
 	case 0x8:	/* Commercial CDs and Audio CD	*/
 		return CD_MEDIA_TYPE_CD;
 	case 0x9:	/* CD-R                         */
@@ -295,10 +314,63 @@ cd_drive_get_media_type (CDDrive *cdrom)
 	return cd_drive_get_media_type_from_path (cdrom->device);
 }
 
+gint64
+cd_drive_get_media_size_from_path (const char *device)
+{
+	int fd;
+	int secs;
+	int mmc_profile;
+	gint64 size;
+
+	g_return_val_if_fail (device != NULL, CD_MEDIA_SIZE_UNKNOWN);
+
+	secs = 0;
+
+	fd = open (device, O_RDONLY|O_EXCL|O_NONBLOCK);
+	if (fd < 0) {
+		if (errno == EBUSY) {
+			return CD_MEDIA_SIZE_BUSY;
+		}
+		return CD_MEDIA_SIZE_UNKNOWN;
+	}
+
+	mmc_profile = get_mmc_profile (fd);
+
+	/* See cd_drive_get_media_type_from_path for details */
+	switch (mmc_profile) {
+	case 0x9:
+	case 0xa:
+		secs = get_disc_size_cd (fd);
+		size = (1 + secs * 7 / 48) * 1024 * 1024;
+		break;
+	case 0x11:
+	case 0x13:
+	case 0x14:
+	case 0x1B:
+	case 0x1A:
+	case 0x12:
+		size = get_disc_size_dvd (fd, mmc_profile);
+		break;
+	default:
+		size = CD_MEDIA_SIZE_NA;
+	}
+
+	close (fd);
+
+	return size;
+}
+
+gint64
+cd_drive_get_media_size (CDDrive *cdrom)
+{
+	g_return_val_if_fail (cdrom != NULL, CD_MEDIA_SIZE_UNKNOWN);
+
+	return cd_drive_get_media_size_from_path (cdrom->device);
+}
+
 #ifdef USE_HAL
 
 #define GET_BOOL_PROP(x) (hal_device_property_exists (ctx, device_names[i], x) && hal_device_get_property_bool (ctx, device_names[i], x))
-#define CD_ROM_SPEED 176
 
 static GList *
 hal_scan (gboolean recorder_only)
@@ -322,7 +394,9 @@ hal_scan (gboolean recorder_only)
 		return NULL;
 	}
 
-	device_names = hal_get_all_devices (ctx, &num_devices);
+	device_names = hal_find_device_by_capability (ctx,
+			"storage.cdrom", &num_devices);
+
 	if (device_names == NULL)
 	{
 		hal_shutdown (ctx);
@@ -335,52 +409,33 @@ hal_scan (gboolean recorder_only)
 		char *string;
 		gboolean is_cdr;
 
-		/* Is it a removable drive? */
-		if (!GET_BOOL_PROP ("storage.removable")) {
-			continue;
-		}
-
-		/* Is it a CD drive? */
-		string = hal_device_get_property_string (ctx, device_names[i],
-				"storage.media");
-		if (string == NULL || strcmp (string, "cdrom") != 0) {
-			g_free (string);
-			continue;
-		}
-		g_free (string);
-
-		/* Is it the CD-drive, or a volume on the media? */
-		if (GET_BOOL_PROP ("block.is_volume") != FALSE) {
-			continue;
-		}
-
 		/* Is it a CD burner? */
-		is_cdr = GET_BOOL_PROP ("storage.cdr");
+		is_cdr = GET_BOOL_PROP ("storage.cdrom.cdr");
 
-		cdrom = g_new0 (CDDrive, 1);
+		cdrom = cd_drive_new ();
 		cdrom->type = CDDRIVE_TYPE_CD_DRIVE;
 		if (is_cdr != FALSE) {
 			cdrom->type |= CDDRIVE_TYPE_CD_RECORDER;
 		}
-		if (GET_BOOL_PROP ("storage.cdrw")) {
+		if (GET_BOOL_PROP ("storage.cdrom.cdrw")) {
 			cdrom->type |= CDDRIVE_TYPE_CDRW_RECORDER;
 		}
-		if (GET_BOOL_PROP ("storage.dvd")) {
+		if (GET_BOOL_PROP ("storage.cdrom.dvd")) {
 			cdrom->type |= CDDRIVE_TYPE_DVD_DRIVE;
 
-			if (GET_BOOL_PROP ("storage.dvdram")) {
+			if (GET_BOOL_PROP ("storage.cdrom.dvdram")) {
 				cdrom->type |= CDDRIVE_TYPE_DVD_RAM_RECORDER;
 			}
-			if (GET_BOOL_PROP ("storage.dvdr")) {
+			if (GET_BOOL_PROP ("storage.cdrom.dvdr")) {
 				cdrom->type |= CDDRIVE_TYPE_DVD_RW_RECORDER;
 			}
-			if (GET_BOOL_PROP ("storage.dvd")) {
+			if (GET_BOOL_PROP ("storage.cdrom.dvd")) {
 				cdrom->type |= CDDRIVE_TYPE_DVD_DRIVE;
 			}
-			if (GET_BOOL_PROP ("storage.dvdplusr")) {
+			if (GET_BOOL_PROP ("storage.cdrom.dvdplusr")) {
 				cdrom->type |= CDDRIVE_TYPE_DVD_PLUS_R_RECORDER;
 			}
-			if (GET_BOOL_PROP ("storage.dvdplusrw")) {
+			if (GET_BOOL_PROP ("storage.cdrom.dvdplusrw")) {
 				cdrom->type |= CDDRIVE_TYPE_DVD_PLUS_RW_RECORDER;
 			}
 		}
@@ -394,7 +449,7 @@ hal_scan (gboolean recorder_only)
 		if (string != NULL) {
 			cdrom->display_name = string;
 		} else {
-			cdrom->display_name = g_strdup_printf ("Unnamed CD-ROM (%s)", cdrom->device);
+			cdrom->display_name = g_strdup_printf ("Unnamed Drive (%s)", cdrom->device);
 		}
 
 		cdrom->max_speed_read = hal_device_get_property_int
@@ -415,8 +470,13 @@ hal_scan (gboolean recorder_only)
 		} else {
 			cdroms = g_list_prepend (cdroms, cdrom);
 		}
+
+		cdrom->priv->udi = g_strdup (device_names[i]);
 	}
 
+	g_strfreev (device_names);
+	hal_shutdown (ctx);
+
 	cdroms = g_list_reverse (cdroms);
 
 	return cdroms;
@@ -629,32 +689,46 @@ lookup_scsi_unit (int bus, int id, int lun,
 	return NULL;
 }
 
+#if !defined(__linux)
 static int
-get_device_max_speed (char *id)
+get_device_max_read_speed (char *device)
 {
+	int fd;
 	int max_speed;
-	char *stdout_data, *speed;
+	int read_speed, write_speed;
 
 	max_speed = -1;
 
-	stdout_data = cdrecord_get_stdout_for_id (id);
-	if (stdout_data == NULL) {
-		return max_speed;
+	fd = open (device, O_RDONLY|O_EXCL|O_NONBLOCK);
+	if (fd < 0) {
+		return -1;
 	}
 
-	speed = strstr (stdout_data, "Maximum write speed in kB/s:");
-	if (speed != NULL) {
-		speed += strlen ("Maximum write speed in kB/s:");
-		max_speed = (int)floor (atol (speed) / 176.0 + 0.5);
-	} else {
-		speed = strstr (stdout_data, "Maximum write speed:");
-		if (speed != NULL) {
-			speed += strlen ("Maximum write speed:");
-			max_speed = (int)floor (atol (speed) / 176.0 + 0.5);
-		}
+	get_read_write_speed (fd, &read_speed, &write_speed);
+	close (fd);
+	max_speed = (int)floor  (read_speed) / CD_ROM_SPEED;
+
+	return max_speed;
+}
+#endif
+
+static int
+get_device_max_write_speed (char *device)
+{
+	int fd;
+	int max_speed;
+	int read_speed, write_speed;
+
+	max_speed = -1;
+
+	fd = open (device, O_RDONLY|O_EXCL|O_NONBLOCK);
+	if (fd < 0) {
+		return -1;
 	}
 
-	g_free (stdout_data);
+	get_read_write_speed (fd, &read_speed, &write_speed);
+	close (fd);
+	max_speed = (int)floor  (write_speed) / CD_ROM_SPEED;
 
 	return max_speed;
 }
@@ -667,7 +741,7 @@ get_scsi_cd_name (int bus, int id, int lun, const char *dev,
 
 	scsi_unit = lookup_scsi_unit (bus, id, lun, scsi_units, n_scsi_units);
 	if (scsi_unit == NULL) {
-		return g_strdup_printf (_("Unnamed SCSI CD-ROM (%s)"), dev);
+		return g_strdup_printf (_("Unnamed SCSI Drive (%s)"), dev);
 	}
 
 	return g_strdup_printf ("%s - %s",
@@ -732,13 +806,13 @@ add_linux_cd_recorder (GList *cdroms,
 	int bus, id, lun;
 	CDDrive *cdrom;
 
-	cdrom = g_new0 (CDDrive, 1);
+	cdrom = cd_drive_new ();
 
 	cdrom->type = CDDRIVE_TYPE_CD_DRIVE;
 	cdrom->display_name = g_strdup (cdrom_s->display_name);
 
 	if (cdrom_s->protocol == CD_PROTOCOL_SCSI) {
-		cdrom->protocol = CD_PROTOCOL_SCSI;
+		cdrom->priv->protocol = CD_PROTOCOL_SCSI;
 		if (!get_cd_scsi_id (cdrom_s->device, &bus, &id, &lun)) {
 			g_free (cdrom->display_name);
 			g_free (cdrom);
@@ -747,15 +821,15 @@ add_linux_cd_recorder (GList *cdroms,
 		cdrom->cdrecord_id = g_strdup_printf ("%d,%d,%d",
 				bus, id, lun);
 	} else {
-		cdrom->protocol = CD_PROTOCOL_IDE;
+		cdrom->priv->protocol = CD_PROTOCOL_IDE;
 		/* kernel >=2.5 can write cd w/o ide-scsi */
 		cdrom->cdrecord_id = g_strdup_printf ("/dev/%s",
 				cdrom_s->device);
 	}
 
 	if (recorder_only) {
-		cdrom->max_speed_write = get_device_max_speed
-			(cdrom->cdrecord_id);
+		cdrom->max_speed_write = get_device_max_write_speed
+			(cdrom->device);
 		if (cdrom->max_speed_write == -1) {
 			cdrom->max_speed_write = cdrom_s->speed;
 		}
@@ -794,7 +868,7 @@ add_linux_cd_drive (GList *cdroms, struct cdrom_unit *cdrom_s,
 {
 	CDDrive *cdrom;
 
-	cdrom = g_new0 (CDDrive, 1);
+	cdrom = cd_drive_new ();
 	cdrom->type = CDDRIVE_TYPE_CD_DRIVE;
 	cdrom->cdrecord_id = NULL;
 	cdrom->display_name = g_strdup (cdrom_s->display_name);
@@ -1075,14 +1149,14 @@ freebsd_scan (gboolean recorder_only)
 			continue;
 		}
 
-		cdrom = g_new0 (CDDrive, 1);
+		cdrom = cd_drive_new ();
 		cdrom->display_name = g_strdup_printf ("%s %s", cam_dev->inq_data.vendor, cam_dev->inq_data.revision);
 		cdrom->device = g_strdup (cam_path);
 		cdrom->cdrecord_id = g_strdup_printf ("%d,%d,%d", cam_dev->path_id, cam_dev->target_id, cam_dev->target_lun);
 		/* Attempt to get more specific information from
 		 * this drive by using cdrecord.
 		 */
-		get_cd_properties (cdrom->cdrecord_id,
+		get_cd_properties (cdrom->device, cdrom->cdrecord_id,
 			&(cdrom->max_speed_read),
 			&(cdrom->max_speed_write),
 			&(cdrom->type));
@@ -1179,14 +1253,14 @@ cdrecord_scan (gboolean recorder_only)
 
 		vendor[8] = '\0'; model[16] = '\0';
 
-		cdrom = g_new0 (CDDrive, 1);
+		cdrom = cd_drive_new ();
 		cdrom->display_name = g_strdup_printf ("%s - %s",
 				g_strstrip (vendor), g_strstrip (model));
 		cdrom->cdrecord_id = g_strdup_printf ("%d,%d,%d", bus, id, lun);
 		/* FIXME we don't have any way to guess the real device
 		 * from the info we get from CDRecord */
-		cdrom->device = g_strdup_printf ("/dev/%d", index);
-		get_cd_properties (cdrom->cdrecord_id,
+		cdrom->device = g_strdup_printf ("/dev/pg%d", index);
+		get_cd_properties (cdrom->device, cdrom->cdrecord_id,
 			&(cdrom->max_speed_read),
 			&(cdrom->max_speed_write),
 			&(cdrom->type));
@@ -1226,7 +1300,7 @@ cd_drive_get_file_image (void)
 {
 	CDDrive *cdrom;
 
-	cdrom = g_new0 (CDDrive, 1);
+	cdrom = cd_drive_new ();
 	cdrom->display_name = g_strdup (_("File image"));
 	cdrom->max_speed_read = 0;
 	cdrom->max_speed_write = 0;
@@ -1268,9 +1342,23 @@ cd_drive_free (CDDrive *drive)
 {
 	g_return_if_fail (drive != NULL);
 
+	if (drive->priv) {
+		g_free (drive->priv->udi);
+		g_free (drive->priv);
+	}
+
 	g_free (drive->display_name);
 	g_free (drive->cdrecord_id);
 	g_free (drive->device);
 	g_free (drive);
 }
 
+static CDDrive *
+cd_drive_new (void)
+{
+	CDDrive *cdrom;
+
+	cdrom = g_new0 (CDDrive, 1);
+	cdrom->priv = g_new0 (CDDrivePriv, 1);
+	return cdrom;
+}
